@@ -4,7 +4,11 @@
 #include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
+#if USE_MONOTONIC_CLOCK
+#include <time.h>
+#else
 #include <sys/time.h>
+#endif
 
 #ifndef X_DISPLAY_MISSING
 #define X_DISPLAY_MISSING
@@ -16,14 +20,18 @@
 static char         progress_called;
 static FILE        *fout;
 
+#define Vprintf(fmt...) if (verbose) fprintf(fout, fmt)
+
 #define HELP \
    "Usage:\n" \
    "  imlib2_load [OPTIONS] FILE...\n" \
    "OPTIONS:\n" \
    "  -e  : Break on error\n" \
    "  -f  : Load with imlib_load_image_fd()\n" \
+   "  -i  : Load with imlib_load_image_immediately()\n" \
    "  -n N: Reeat load N times\n" \
    "  -p  : Check that progress is called\n" \
+   "  -v  : Increase verbosity\n" \
    "  -x  : Print to stderr\n"
 
 static void
@@ -84,20 +92,24 @@ main(int argc, char **argv)
    Imlib_Image         im;
    Imlib_Load_Error    lerr;
    unsigned int        t0;
+   int                 verbose;
    int                 check_progress;
    int                 break_on_error;
    int                 show_time;
    int                 load_cnt, cnt;
    int                 load_fd;
+   int                 load_now;
 
    fout = stdout;
+   verbose = 0;
    check_progress = 0;
    break_on_error = 0;
-   load_cnt = 1;
    show_time = 0;
+   load_cnt = 1;
    load_fd = 0;
+   load_now = 0;
 
-   while ((opt = getopt(argc, argv, "efn:px")) != -1)
+   while ((opt = getopt(argc, argv, "efin:pvx")) != -1)
      {
         switch (opt)
           {
@@ -107,12 +119,20 @@ main(int argc, char **argv)
           case 'f':
              load_fd = 1;
              break;
+          case 'i':
+             load_now = 1;
+             break;
           case 'n':
              load_cnt = atoi(optarg);
              show_time = 1;
+             verbose = 1;
              break;
           case 'p':
              check_progress = 1;
+             verbose = 1;
+             break;
+          case 'v':
+             verbose = 1;
              break;
           case 'x':
              fout = stderr;
@@ -141,7 +161,7 @@ main(int argc, char **argv)
      {
         progress_called = 0;
 
-        fprintf(fout, "Loading image: '%s'\n", argv[0]);
+        Vprintf("Loading image: '%s'\n", argv[0]);
 
         if (show_time)
            t0 = time_us();
@@ -154,6 +174,8 @@ main(int argc, char **argv)
                 im = imlib_load_image_with_error_return(argv[0], &lerr);
              else if (load_fd)
                 im = image_load_fd(argv[0]);
+             else if (load_now)
+                im = imlib_load_image_immediately(argv[0]);
              else
                 im = imlib_load_image(argv[0]);
 
@@ -168,7 +190,7 @@ main(int argc, char **argv)
 
              imlib_context_set_image(im);
 
-             if (!check_progress)
+             if (!check_progress && !load_now)
                 imlib_image_get_data();
 
              imlib_free_image_and_decache();
