@@ -1,4 +1,5 @@
 #include "loader_common.h"
+
 #include <bzlib.h>
 
 #define OUTBUF_SIZE 16384
@@ -91,6 +92,5 @@ void
 formats(ImlibLoader * l)
 {
    static const char  *const list_formats[] = { "bz2" };
-   __imlib_LoaderSetFormats(l, list_formats,
-                            sizeof(list_formats) / sizeof(char *));
+   __imlib_LoaderSetFormats(l, list_formats, ARRAY_SIZE(list_formats));
 }
