@@ -1,4 +1,5 @@
 #include "loader_common.h"
+
 #include <gif_lib.h>
 
 int
@@ -104,10 +105,7 @@ load2(ImlibImage * im, int load_data)
      }
    while (rec != TERMINATE_RECORD_TYPE);
 
-   if (transp >= 0)
-      SET_FLAG(im->flags, F_HAS_ALPHA);
-   else
-      UNSET_FLAG(im->flags, F_HAS_ALPHA);
+   UPDATE_FLAG(im->flags, F_HAS_ALPHA, transp >= 0);
 
    if (!rows)
       goto quit;
@@ -182,6 +180,5 @@ void
 formats(ImlibLoader * l)
 {
    static const char  *const list_formats[] = { "gif" };
-   __imlib_LoaderSetFormats(l, list_formats,
-                            sizeof(list_formats) / sizeof(char *));
+   __imlib_LoaderSetFormats(l, list_formats, ARRAY_SIZE(list_formats));
 }
