@@ -1,4 +1,5 @@
 #include "loader_common.h"
+
 #include <zlib.h>
 
 #define OUTBUF_SIZE 16484
@@ -84,6 +85,5 @@ void
 formats(ImlibLoader * l)
 {
    static const char  *const list_formats[] = { "gz" };
-   __imlib_LoaderSetFormats(l, list_formats,
-                            sizeof(list_formats) / sizeof(char *));
+   __imlib_LoaderSetFormats(l, list_formats, ARRAY_SIZE(list_formats));
 }
